
from __future__ import annotations

from abc import abstractmethod, ABCMeta
from argparse import Namespace, ArgumentParser
from collections.abc import Sequence
from contextlib import AsyncExitStack
from typing import Protocol, Any

from .login import LoginInterface
from ..config import IMAPConfig
from ..health import HealthStatus

__all__ = ['BackendInterface', 'ServiceInterface']


class BackendInterface(Protocol):
    """Defines the abstract base class that is expected for backends that
    register themselves on the ``pymap.backend`` entry point.

    """

    __slots__: Sequence[str] = []

    @classmethod
    @abstractmethod
    def add_subparser(cls, name: str, subparsers: Any) -> ArgumentParser:
        """Add a command-line argument sub-parser that will be used to choose
        this backend. For example::

            parser = subparsers.add_parser('foo', help='foo backend')
            parser.add_argument(...)

        Args:
            name: The name to use for the subparser.
            subparsers: The special action object as returned by
                :meth:`~argparse.ArgumentParser.add_subparsers`.

        Returns:
            The new sub-parser object.

        """
        ...

    @classmethod
    @abstractmethod
    async def init(cls, args: Namespace, **overrides: Any) \
            -> tuple[BackendInterface, IMAPConfig]:
        """Initialize the backend and return an instance.

        Args:
            args: The command-line arguments.
            overrides: Override keyword arguments to the config constructor.

        """
        ...

    @abstractmethod
    async def start(self, stack: AsyncExitStack) -> None:
        """Start the backend.

        Args:
            stack: An exit stack that should be used for cleanup.

        """
        ...

    @property
    @abstractmethod
    def login(self) -> LoginInterface:
        """Login interface that handles authentication credentials."""
        ...

    @property
    @abstractmethod
    def config(self) -> IMAPConfig:
        """The IMAP config in use by the backend."""
        ...

    @property
    @abstractmethod
    def status(self) -> HealthStatus:
        """The health status for the backend."""
        ...


class ServiceInterface(metaclass=ABCMeta):
    """Defines the abstract base class that is expected for services that
    register themselves on the ``pymap.service`` entry point.

    """

    __slots__ = ['backend', 'config']

    def __init__(self, backend: BackendInterface, config: IMAPConfig) -> None:
        super().__init__()
        self.backend = backend
        self.config = config

    @classmethod
    @abstractmethod
    def add_arguments(cls, parser: ArgumentParser) -> None:
        """Add the arguments or argument group used to configure the service.
        For example::

            group = parser.add_argument_group('foo service arguments')
            group.add_argument(...)

        Args:
            parser: The argument parser.

        """
        ...

    @abstractmethod
    async def start(self, stack: AsyncExitStack) -> None:
        """Start the service.

        Args:
            stack: An exit stack that should be used for cleanup.

        """
        ...




from __future__ import annotations

import asyncio
import logging
from argparse import ArgumentParser, Namespace
from collections.abc import Mapping
from contextlib import AsyncExitStack

from pymap.context import cluster_metadata
from pymap.interfaces.backend import ServiceInterface
from swimprotocol.config import ConfigError, TransientConfigError
from swimprotocol.members import Member, Members
from swimprotocol.status import Status
from swimprotocol.transport import load_transport


__all__ = ['transport_type', 'SwimService']

_log = logging.getLogger(__name__)

#: The :class:`~swimprotocol.transport.Transport` implementation.
transport_type = load_transport()


class SwimService(ServiceInterface):  # pragma: no cover
    """A pymap service implemented using `swim-protocol
    <https://icgood.github.io/swim-protocol/>`_ to establish a cluster of pymap
    instances.

    """

    @classmethod
    def add_arguments(cls, parser: ArgumentParser) -> None:
        transport_type.config_type.add_arguments(parser, prefix='--swim-')

    async def _remote_update(self, member: Member) -> None:
        if member.status & Status.AVAILABLE \
                and member.metadata is not Member.METADATA_UNKNOWN:
            cluster_metadata.get().add(member)
        else:
            cluster_metadata.get().discard(member)

    def _local_update(self, members: Members,
                      metadata: Mapping[str, bytes]) -> None:
        members.update(members.local, new_metadata=metadata)

    async def _start(self, args: Namespace, stack: AsyncExitStack) -> None:
        while True:
            try:
                config = transport_type.config_type.from_args(
                    args, ping_interval=10.0,
                    ping_timeout=1.0,
                    suspect_timeout=300.0,
                    sync_interval=30.0)
            except TransientConfigError as exc:
                _log.debug('SWIM configuration failure: %s', exc)
                await asyncio.sleep(exc.wait_hint)
            except ConfigError:
                return  # do not run SWIM if not configured properly
            else:
                break
        _log.debug('SWIM configuration: %r %r',
                   config.local_name, config.peers)

        transport = transport_type(config)
        members = Members(config)
        cluster_metadata.get().listen(self._local_update, members)
        stack.enter_context(members.listener.on_notify(self._remote_update))
        worker = await stack.enter_async_context(transport.enter(members))
        task = asyncio.create_task(worker.run())
        stack.callback(task.cancel)

    async def start(self, stack: AsyncExitStack) -> None:
        args = self.config.args
        swim_stack = AsyncExitStack()
        await stack.enter_async_context(swim_stack)
        task = asyncio.create_task(self._start(args, swim_stack))
        stack.callback(task.cancel)




from collections.abc import Iterable, Sequence
from typing import Union

__all__ = ['whitespace', 'find_any', 'get_raw']

whitespace = frozenset(b' \t\n\r\x0b\f')

_Line = tuple[int, int, int]
_Lines = Sequence[_Line]


def find_any(data: Union[bytes, memoryview], end_marker: frozenset[int],
             start: int, end: int, inverse: bool, reverse: bool) -> int:
    if reverse:
        range_iter: Iterable[int] = reversed(range(start, end))
    else:
        range_iter = range(start, end)
    for i in range_iter:
        match = (data[i] in end_marker)
        if (not match and not inverse) or (match and inverse):
            if reverse:
                return i + 1
            else:
                return i
    return -1


def get_raw(view: memoryview, *lines: _Lines) -> memoryview:
    try:
        start = lines[0][0][0]
    except IndexError:
        start = 0
    try:
        end = lines[-1][-1][2]
    except IndexError:
        end = -1
    return view[start:end]




from __future__ import annotations

import asyncio
import logging
import time
from typing import ClassVar

from aioredis import Redis

from .background import BackgroundAction
from .keys import GlobalKeys, CleanupKeys, NamespaceKeys, ContentKeys, \
    MailboxKeys
from .scripts.cleanup import CleanupScripts

__all__ = ['CleanupAction']

_log = logging.getLogger(__name__)
_scripts = CleanupScripts()


class CleanupAction(BackgroundAction):
    """Defines the logic for monitoring and executing cleanup of various
    entities.

    Args:
        address: The redis server address.
        redis: The redis connection object.
        global_keys: The global keys group.

    """

    namespace_ttl: ClassVar[int] = 0
    mailbox_ttl: ClassVar[int] = 600
    content_ttl: ClassVar[int] = 3600

    def __init__(self, global_keys: GlobalKeys) -> None:
        super().__init__()
        self._global_keys = global_keys
        self._keys = keys = CleanupKeys(global_keys)
        self._order = (keys.mailboxes, keys.namespaces, keys.contents)

    async def __call__(self, conn: Redis, duration: float) -> None:
        now = time.time()
        finished = now + duration
        while now < finished:
            timeout: int = finished - now  # type: ignore
            cleanup_key, cleanup_val = await conn.blpop(
                self._order, timeout=timeout)
            try:
                await asyncio.shield(
                    self._run_one(conn, cleanup_key, cleanup_val))
            except Exception:
                _log.warning('Cleanup failed: key=%s val=%s',
                             cleanup_key, cleanup_val, exc_info=True)
                raise
            now = time.time()

    async def _run_one(self, conn: Redis, cleanup_key: bytes,
                       cleanup_val: bytes) -> None:
        keys = self._keys
        if cleanup_key == keys.namespaces:
            namespace = cleanup_val
            await self._run_namespace(conn, namespace)
        elif cleanup_key == keys.mailboxes:
            namespace, mailbox_id = cleanup_val.split(b'\x00', 1)
            await self._run_mailbox(conn, namespace, mailbox_id)
        elif cleanup_key == keys.contents:
            namespace, email_id = cleanup_val.split(b'\x00', 1)
            await self._run_content(conn, namespace, email_id)

    async def _run_namespace(self, conn: Redis, namespace: bytes) -> None:
        ns_keys = NamespaceKeys(self._global_keys, namespace)
        await _scripts.namespace(conn, self._keys, ns_keys,
                                 ttl=self.namespace_ttl)

    async def _run_mailbox(self, conn: Redis, namespace: bytes,
                           mailbox_id: bytes) -> None:
        ns_keys = NamespaceKeys(self._global_keys, namespace)
        mbx_keys = MailboxKeys(ns_keys, mailbox_id)
        await _scripts.mailbox(conn, self._keys, mbx_keys,
                               ttl=self.mailbox_ttl)

    async def _run_content(self, conn: Redis, namespace: bytes,
                           email_id: bytes) -> None:
        ns_keys = NamespaceKeys(self._global_keys, namespace)
        ct_keys = ContentKeys(ns_keys, email_id)
        await _scripts.content(conn, ns_keys, ct_keys,
                               ttl=self.content_ttl)



"""Base implementations of the :mod:`pymap.interfaces.message` interfaces."""

from __future__ import annotations

import re
from abc import ABCMeta
from collections.abc import Collection, Iterable, Mapping, Sequence
from datetime import datetime
from typing import Any, Optional, Final

from .bytes import Writeable
from .flags import SessionFlags
from .interfaces.message import FlagsKey, CachedMessage, MessageInterface, \
    LoadedMessageInterface
from .mime import MessageContent
from .mime.cte import MessageDecoder
from .parsing.response.fetch import EnvelopeStructure, BodyStructure, \
    MultipartBodyStructure, ContentBodyStructure, TextBodyStructure, \
    MessageBodyStructure
from .parsing.specials import Flag, ObjectId, FetchRequirement

__all__ = ['BaseMessage', 'BaseLoadedMessage']


class _NoContent(ValueError):
    # Thrown when message contents were requested, but the backend did not or
    # could not load them.

    def __init__(self) -> None:
        super().__init__('Message content not available.')


class BaseMessage(MessageInterface, CachedMessage, metaclass=ABCMeta):
    """Message metadata such as UID, permanent flags, and when the message
    was added to the system.

    Args:
        uid: The UID of the message.
        internal_date: The internal date of the message.
        permanent_flags: Permanent flags for the message.
        email_id: The message content identifier for the message.
        thread_id: The thread identifier for the message.
        expunged: True if this message has been expunged from the mailbox.

    """

    __slots__ = ['uid', 'internal_date', 'expunged', '_permanent_flags',
                 '_email_id', '_thread_id', '_flags_key']

    def __init__(self, uid: int, internal_date: datetime,
                 permanent_flags: Iterable[Flag], *,
                 email_id: ObjectId = None, thread_id: ObjectId = None,
                 expunged: bool = False) -> None:
        super().__init__()
        self.uid: Final = uid
        self.internal_date: Final = internal_date
        self.expunged: Final = expunged
        self._email_id = email_id or ObjectId(None)
        self._thread_id = thread_id or ObjectId(None)
        self._permanent_flags = frozenset(permanent_flags or ())
        self._flags_key = (uid, self._permanent_flags)

    @property
    def email_id(self) -> ObjectId:
        return self._email_id

    @property
    def thread_id(self) -> ObjectId:
        return self._thread_id

    @property
    def permanent_flags(self) -> frozenset[Flag]:
        return self._permanent_flags

    @permanent_flags.setter
    def permanent_flags(self, permanent_flags: frozenset[Flag]) -> None:
        self._permanent_flags = permanent_flags
        self._flags_key = (self.uid, permanent_flags)

    def get_flags(self, session_flags: SessionFlags) -> frozenset[Flag]:
        msg_sflags = session_flags.get(self.uid)
        if msg_sflags:
            return self._permanent_flags | msg_sflags
        else:
            return self._permanent_flags

    @property
    def flags_key(self) -> FlagsKey:
        return self._flags_key

    def __repr__(self) -> str:
        type_name = type(self).__name__
        return f'<{type_name} uid={self.uid} flags={self.permanent_flags}>'


class BaseLoadedMessage(LoadedMessageInterface, metaclass=ABCMeta):
    """The loaded message content, implemented using an instance of
    :class:`~pymap.mime.MessageContent`.

    Args:
        message: The message object.
        requirement: The fetch requirement of the loaded content.
        content: The MIME-parsed message content, if available.

    """

    __slots__ = ['_message', '_requirement', '_content']

    def __init__(self, message: MessageInterface,
                 requirement: FetchRequirement,
                 content: Optional[MessageContent]) -> None:
        super().__init__()
        self._message = message
        self._requirement = requirement
        self._content = content

    @property
    def message(self) -> MessageInterface:
        return self._message

    @property
    def requirement(self) -> FetchRequirement:
        return self._requirement

    @property
    def content(self) -> MessageContent:
        if self._content is None:
            raise _NoContent()
        return self._content

    def __bytes__(self) -> bytes:
        return bytes(self.content)

    def _get_subpart(self, section) -> MessageContent:
        if section:
            subpart = self.content
            for i in section:
                if subpart.body.has_nested:
                    subpart = subpart.body.nested[i - 1]
                elif i == 1:
                    pass
                else:
                    raise IndexError(i)
            return subpart
        else:
            return self.content

    def get_header(self, name: bytes) -> Sequence[str]:
        try:
            return self.content.header.parsed[name]
        except (KeyError, _NoContent):
            return []

    def get_headers(self, section: Sequence[int]) -> Writeable:
        try:
            msg = self._get_subpart(section)
        except (IndexError, _NoContent):
            return Writeable.empty()
        else:
            return msg.header

    def get_body(self, section: Sequence[int] = None,
                 binary: bool = False) -> Writeable:
        try:
            msg = self._get_subpart(section)
        except (IndexError, _NoContent):
            return Writeable.empty()
        if binary:
            decoded = MessageDecoder.of(msg.header).decode(msg.body)
            if not section:
                return Writeable.concat((msg.header, decoded))
            else:
                return decoded
        else:
            if not section:
                return msg
            else:
                return msg.body

    def get_message_headers(self, section: Sequence[int] = None,
                            subset: Collection[bytes] = None,
                            inverse: bool = False) -> Writeable:
        try:
            msg = self._get_subpart(section)
        except (IndexError, _NoContent):
            return Writeable.empty()
        if section:
            if msg.is_rfc822:
                msg = msg.body.nested[0]
            else:
                return Writeable.empty()
        if subset is None:
            return msg.header
        headers = Writeable.concat(value for key, value in msg.header.folded
                                   if inverse != (key.upper() in subset))
        return Writeable.concat((headers, Writeable.wrap(b'\r\n')))

    def get_message_text(self, section: Sequence[int] = None) -> Writeable:
        try:
            msg = self._get_subpart(section)
        except (IndexError, _NoContent):
            return Writeable.empty()
        if section:
            if msg.is_rfc822:
                msg = msg.body.nested[0]
            else:
                return Writeable.empty()
        return msg.body

    @classmethod
    def _get_size_with_lines(cls, msg: MessageContent) -> tuple[int, int]:
        return len(msg), msg.lines

    def get_size(self, section: Sequence[int] = None) -> int:
        try:
            msg = self._get_subpart(section)
        except (IndexError, _NoContent):
            return 0
        return len(msg)

    def get_envelope_structure(self) -> EnvelopeStructure:
        try:
            return self._get_envelope_structure(self.content)
        except _NoContent:
            return EnvelopeStructure.empty()

    def get_body_structure(self) -> BodyStructure:
        try:
            return self._get_body_structure(self.content)
        except _NoContent:
            return BodyStructure.empty()

    @classmethod
    def _get_envelope_structure(cls, msg: MessageContent) -> EnvelopeStructure:
        parsed = msg.header.parsed
        return EnvelopeStructure(
            parsed.date, parsed.subject, parsed.from_, parsed.sender,
            parsed.reply_to, parsed.to, parsed.cc, parsed.bcc,
            parsed.in_reply_to, parsed.message_id)

    @classmethod
    def _get_params(cls, msg: MessageContent) -> Mapping[str, Any]:
        return msg.body.content_type.params

    @classmethod
    def _get_body_structure(cls, msg: MessageContent) -> BodyStructure:
        parsed = msg.header.parsed
        maintype = msg.body.content_type.maintype
        subtype = msg.body.content_type.subtype
        params = cls._get_params(msg)
        disposition = parsed.content_disposition
        language = parsed.content_language
        location = parsed.content_location
        if maintype == 'multipart':
            sub_body_structs = [cls._get_body_structure(part)
                                for part in msg.body.nested]
            return MultipartBodyStructure(
                subtype, params, disposition, language, location,
                sub_body_structs)
        content_id = parsed.content_id
        content_desc = parsed.content_description
        content_encoding = parsed.content_transfer_encoding
        if maintype == 'message' and subtype == 'rfc822':
            sub_msg = msg.body.nested[0]
            sub_env_struct = cls._get_envelope_structure(sub_msg)
            sub_body_struct = cls._get_body_structure(sub_msg)
            size, lines = cls._get_size_with_lines(msg)
            return MessageBodyStructure(
                params, disposition, language, location, content_id,
                content_desc, content_encoding, None, size, lines,
                sub_env_struct, sub_body_struct)
        elif maintype == 'text':
            size, lines = cls._get_size_with_lines(msg)
            return TextBodyStructure(
                subtype, params, disposition, language, location,
                content_id, content_desc, content_encoding, None, size, lines)
        size = len(msg)
        return ContentBodyStructure(
            maintype, subtype, params, disposition, language, location,
            content_id, content_desc, content_encoding, None, size)

    def contains(self, value: bytes) -> bool:
        try:
            content = self.content
        except _NoContent:
            return False
        pattern = re.compile(re.escape(value), re.I)
        for part in content.walk():
            if pattern.search(bytes(part.header)) is not None:
                return True
            elif part.body.content_type.maintype == 'text':
                if pattern.search(bytes(part.body)) is not None:
                    return True
        return False
